/*
 * queue.c
 *
 *  version 0.10  february 3, 1997
 *  jeremy wohl (jeremyw@evantide.com)
 *  no warranty, no responsibility, etc.
 *
 */

#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>

#include "common.h"

typedef struct _node_t {
    struct _node_t * next;
    void           * datum;
} node_t;

typedef struct {
    sem_t             semaphore;
    pthread_mutex_t   mutex;
    node_t          * first, * last;
    int               items;
} queue_t;


void * queue_create ()
{
    queue_t * queue;
    queue = (queue_t *) malloc (sizeof (queue_t));

    if (queue == 0)  return 0;

    queue->first = 0;
    queue->last  = 0;
    queue->items = 0;

    sem_init (&queue->semaphore, 0, 0);
    pthread_mutex_init (&queue->mutex, 0);

    return ((void *) queue);
}


void queue_free (void * q)
{
    queue_t * queue = (queue_t *) q;
    sem_destroy (&queue->semaphore);
    pthread_mutex_destroy (&queue->mutex);
}


int queue_items (void * q)
{
    queue_t * queue = (queue_t *) q;
    return (queue->items);
}


void queue_lock (void * q)
{
    queue_t * queue = (queue_t *) q;
    pthread_mutex_lock (&queue->mutex);
}


void queue_unlock (void * q)
{
    queue_t * queue = (queue_t *) q;
    pthread_mutex_unlock (&queue->mutex);
}


void queue_sem_wait (void *q)
{
    queue_t * queue = (queue_t *) q;
    sem_wait (&queue->semaphore);
}


int queue_sem_trywait (void * q)
{
    queue_t * queue = (queue_t *) q;
    return (sem_trywait (&queue->semaphore));
}


void queue_add (void * q, void * datum)
{
    queue_t * queue = (queue_t *) q;
    node_t  * node;


    queue_lock (q);

    while ((node = malloc (sizeof (node_t))) == 0)  nomem_sleep ();

    node->next  = 0;
    node->datum = datum;
    if (queue->last) queue->last->next  = node;
    queue->last = node;

    if (queue->first == 0)
	queue->first = queue->last;

    queue->items++;

    sem_post (&queue->semaphore);

    queue_unlock (q);

    return;
}


void * queue_front (void * q)
{
    queue_t * queue = (queue_t *) q;

    if (queue->first)
	return (queue->first->datum);

    return (0);
}


void queue_remove (void * q)
{
    queue_t * queue = (queue_t *) q;
    node_t  * tmp;

    if (queue->first != 0) {
	if (queue->first == queue->last)
	    queue->last = 0;

	tmp = queue->first->next;
	free (queue->first);
	
	queue->first = tmp;

	queue->items--;
    }

    return;
}
