/*
 * trigger_remd.c
 *
 *  version 0.41  april 21, 1998
 *  jeremy wohl (jeremyw@evantide.com)
 *  no warranty, no responsibility, etc.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include "triggermsg.h"

#define MAX_NS 2000

int  read_ns (char * buf, int buf_len);
int  parse_sz (char * pos, char * buf, int buf_len, char ** new_pos);
int  parse_int (char * pos, int * dest, char ** new_pos);
void print_ns (char * string);

void bad_version () { print_ns ("Dunknown protocol version");  exit (0); }
void bad_ns ()      { print_ns ("Dbad netstring");             exit (0); }
void temp_fail ()   { print_ns ("Ztemporary failure");         exit (0); }
void perm_fail ()   { print_ns ("Dpermanent failure");         exit (0); }


int main ()
{
    struct trigger_msg  t;

    FILE  * fp;
    char    ns [MAX_NS], * pos = ns, str [10];
    int     result, msqid;


    if (read_ns (ns, MAX_NS))   bad_ns ();

    if (parse_int (pos, &t.version, &pos))  bad_ns ();
    if (t.version != 1)                     bad_version ();

    if (parse_sz (pos, t.user,    sizeof t.user,    &pos))   bad_ns ();
    if (parse_sz (pos, t.service, sizeof t.service, &pos))   bad_ns ();
    if (parse_sz (pos, t.action,  sizeof t.action,  &pos))   bad_ns ();
    if (parse_sz (pos, t.data,    sizeof t.data,    &pos))   bad_ns ();

    if (parse_int (pos, &t.time_sec,  &pos))   bad_ns ();
    if (parse_int (pos, &t.time_usec, &pos))   bad_ns ();
    if (parse_int (pos, &t.delay,     &pos))   bad_ns ();
    if (parse_int (pos, &t.hops,      0))      bad_ns ();

    t.hops++;

    t.mtype = 1;

    if ((fp = fopen ("/var/run/triggerd.msqid", "r")) == 0)  temp_fail ();
    if (      fgets (str, 10, fp)                     == 0)  temp_fail ();
    fclose (fp);

    msqid = atoi (str);

    result = msgsnd (msqid, (struct msgbuf *) &t, sizeof t, IPC_NOWAIT);

    if (result == -1 && (errno == EAGAIN || errno == EINTR))  temp_fail ();
    if (result == -1)                                         perm_fail ();

    /* @@ temp failure if triggerd isn't running? log it? */
    
    print_ns ("Kok");
    exit (0);
}


/* returns: 0 ok, 1 bad netstring or i/o failure */
int read_ns (char * buf, int buf_len)
{
    int len;


    if (scanf ("%9d", &len) < 1)   return (1);
    if (len > buf_len - 1)         return (1);

    if (getchar () != ':')         return (1);

    if (fread (buf, 1, len, stdin) < len)  return (1);
    buf [len] = 0;

    if (getchar () != ',')         return (1);

    return (0);
}


/* returns: 0 ok, 1 bad netstring */
int parse_sz (char * pos, char * buf, int buf_len, char ** new_pos)
{
    char * tmp_pos;
    int    len;


    len = strtol (pos, &tmp_pos, 10);

    if (len < 0 || len > buf_len - 1)  return (1);
    if (*tmp_pos != ':')               return (1);

    tmp_pos++;
    memcpy (buf, tmp_pos, len);
    buf [len] = 0;

    tmp_pos += len;
    if (*tmp_pos != ',')  return (1);

    if (new_pos)  *new_pos = tmp_pos + 1;

    return (0);
}


/* returns: 0 ok, 1 bad netstring */
int parse_int (char * pos, int * dest, char ** new_pos)
{
    char * tmp_pos;
    int    len;


    len = strtol (pos, &tmp_pos, 10);

    if (len == LONG_MIN || len == LONG_MAX)  return (1);
    if (*tmp_pos != ':')                     return (1);

    tmp_pos++;
    *dest = strtol (tmp_pos, &tmp_pos, 10);

    if (*dest == LONG_MIN || *dest == LONG_MAX)  return (1);
    if (*tmp_pos != ',')                         return (1);

    if (new_pos)  *new_pos = tmp_pos + 1;

    return (0);
}


void print_ns (char * string)
{
    printf ("%i:%s,", strlen (string), string);
}
